import sys
import re
import logging
import gnupg
import twisted.internet.reactor
import global_config
import keyannounce
import flatstore
import filestore
import keyannounce
import postmessage
import fetchmail

re_keyid = re.compile("^[TBC]:[0123456789abcdef]{40}$",re.IGNORECASE)
re_keyid_ka = re.compile("^[0123456789abcdef]{40}$",re.IGNORECASE)

class cli:

	def main(self,argv):
		cmdline = argv[1:]
		if len(cmdline) == 0:
			#DBGOUT#print "Usage: python cli.py -homedir /path -pubkey [optional keyid]"
			#DBGOUT#print "Usage: python cli.py -homedir /path [-from-entangled] -getkeys hex-keyid hex-keyid"
			#DBGOUT#print "Usage: python cli.py -homedir /path -getkeys email-address"
			#DBGOUT#print "Usage: python cli.py -homedir /path -encode T:hex-keyid C:hex-keyid B:hex-keyid -subject Subject -attachments /path/file1 /path/file2 -content /path/to/content"
			#DBGOUT#print "Usage: python cli.py -homedir /path -push"
			#DBGOUT#print "Usage: python cli.py -homedir /path -fetch"
			#DBGOUT#print "Usage: python cli.py -homedir /path -reassemble hex-announceid dest-path"
			sys.exit(1)
		
		publish_key = False
		get_keys = False
		push_messages = False
		encode_message = False
		self.fetch_from_entangled = False
		fetch_messages = False
		reassemble = False
		self.timeout = 15
		self.keyid_to_publish = None
		homedir = None
		send_to = [ ]
		attachments = [ ]

		logging.basicConfig(level=logging.DEBUG, 
        	format="%(asctime)s %(levelname)-5s %(name)-10s %(threadName)-10s %(message)s")
		logger = logging.getLogger(__name__)
		
		n = 0
		while n < len(cmdline):
			cmd = cmdline[n]
			#DBGOUT#print n,cmd
		
			if cmd == '-homedir':
				n += 1
				homedir = cmdline[n]
				n += 1
			elif cmd == '-pubkey':
				n += 1
				publish_key = True
				if n < len(cmdline) and re_keyid_ka.match(cmdline[n]):
					self.keyid_to_publish = cmdline[n]
					n += 1
			elif cmd == '-from-entangled':
				self.fetch_from_entangled = True
				n += 1
			elif cmd == '-getkeys':
				n += 1
				self.get_keyids = [ ]
				self.get_keyids.append(cmdline[n])
				n += 1
				while n < len(cmdline) and re_keyid.match(cmdline[n]):
					self.get_keyids.append(cmdline[n])
					n += 1
				get_keys = True
			elif cmd == '-encode':
				encode_message = True
				n += 1
				while n < len(cmdline) and re_keyid.match(cmdline[n]):
					send_to.append(cmdline[n])
					n += 1
			elif cmd == '-attachments':
				n += 1
				while n < len(cmdline) and cmdline[n][0] != '-':
					attachments.append(cmdline[n])
					n += 1
			elif cmd == '-subject':
				n += 1
				subject = cmdline[n]
				n += 1
			elif cmd == '-content':
				n += 1
				content = cmdline[n]
				n += 1
			elif cmd == '-push':
				push_messages = True
				n += 1
			elif cmd == '-fetch':
				fetch_messages = True
				n += 1
			elif cmd == '-reassemble':
				reassemble = True
				n += 1
				reassemble_hash = cmdline[n]
				n += 1
				reassemble_dest_path = cmdline[n]
				n += 1
			else:
				#DBGOUT#print "Unrecognized: "+cmd
				n += 1
		
		config_file_path = homedir + global_config.separator + 'config.txt'
		gpg_homedir = homedir + global_config.separator + 'gpg'
		config_file_handle = open(config_file_path,'r')
		config_data = config_file_handle.read()
		config_file_handle.close()
		self.gpg = gnupg.GPG(verbose = True,gnupghome = gpg_homedir)
		self.passphrase = None
		
		
		for line in config_data.split('\n'):
			line = line.rstrip('\r')
			lineL = line.lower()
			#DBGOUT#print lineL
			if lineL[0:7] == 'keyid: ' :
				self.client_keyid_hex = line[7:]
				self.client_keyid = self.client_keyid_hex.decode("hex")
				if self.keyid_to_publish == None:
					self.keyid_to_publish = self.client_keyid_hex
			elif lineL[0:19] == 'senderproofofwork: ':
				self.sender_proof_of_work = line[19:]
			elif lineL[0:11] == 'mailboxes: ':
				self.mailboxes = line[11:]
			elif lineL[0:11] == 'transport: ':
				self.transport = line[11:]
			elif lineL[0:17] == 'entangledserver: ':
				self.entangled_server = line[17:]
		
		#DBGOUT#print 'homedir',homedir
		#DBGOUT#print 'client_keyid',self.client_keyid_hex
		#DBGOUT#print 'sender_proof_of_work',self.sender_proof_of_work
		#DBGOUT#print 'mailboxes',self.mailboxes
		#DBGOUT#print 'transport',self.transport
		#DBGOUT#print 'entangled_server',self.entangled_server
		
		self.local_store = filestore.filestore(homedir + global_config.separator + "localstore")
		self.outbox_store = flatstore.flatstore(homedir + global_config.separator + "outbox")
		self.complete_store = flatstore.flatstore(homedir + global_config.separator + "complete")
		self.incomplete_store = flatstore.flatstore(homedir + global_config.separator + "incomplete")
		poster = None
		if encode_message or push_messages:
			self.poster = postmessage.postmessage(self.gpg,self.outbox_store,self.entangled_server,self.timeout,global_config.separator)
			self.poster.set_max_blocksize(65536,65536)

		if publish_key:
			pow_nbits = 16
			pow_nmatches = 2
			self.keyannounce = keyannounce.keyannounce(self.gpg)
			self.key_announcement,self.email_claim_hash,self.email_claim = self.keyannounce.create_key_announcement_message(self.keyid_to_publish,self.sender_proof_of_work,pow_nbits,pow_nmatches,self.mailboxes,self.transport,self.passphrase)
			self.local_store.store(self.keyid_to_publish,self.key_announcement)
			if self.email_claim_hash != None:
				self.local_store.store(self.email_claim_hash,self.email_claim)
			twisted.internet.reactor.callLater(1,self.run_post_key_announcement)
			twisted.internet.reactor.run()
			return

		elif get_keys:
			self.keyannounce = keyannounce.keyannounce(self.gpg)
			twisted.internet.reactor.callLater(1,self.run_get_key_announcement)
			twisted.internet.reactor.run()
			return

		elif encode_message:
			self.poster.encode_message(self.local_store,content,self.client_keyid,self.transport,send_to,attachments,subject,self.passphrase,self.status_callback)

		elif push_messages:
			twisted.internet.reactor.callLater(1,self.run_push_messages)
			twisted.internet.reactor.run()
			return
		
		elif fetch_messages:
			self.fetcher = fetchmail.fetchmail(self.gpg,self.local_store,self.complete_store,self.incomplete_store,self.entangled_server,self.timeout)
			twisted.internet.reactor.callLater(1,self.run_fetch_messages)
			twisted.internet.reactor.run()

		elif reassemble:
			self.fetcher = fetchmail.fetchmail(self.gpg,self.local_store,self.complete_store,self.incomplete_store,self.entangled_server,self.timeout)
			result,reason = self.fetcher.reassemble_message(reassemble_hash,reassemble_dest_path)
			if result == True:
				#DBGOUT#print "Reassembly successful"
			else:
				#DBGOUT#print "Reassembly failed: "+reason
		
	def run_post_key_announcement(self):
		if self.transport.lower() == 'entangled':
			server = self.entangled_server
		else:
			server = self.transport
		self.keyannounce.post_key_announcement(server,self.keyid_to_publish,self.key_announcement,self.email_claim_hash,self.email_claim,True,True,self.post_key_announcement_callback,self.timeout)

	def post_key_announcement_callback(self,posted_key_to_server,posted_key_to_entangled,posted_claim_to_server,posted_claim_to_entangled):
		#DBGOUT#print "Post key announcement done"
		#DBGOUT#print "posted key to server: ",posted_key_to_server
		#DBGOUT#print "posted key to entangled: ",posted_key_to_entangled
		#DBGOUT#print "posted claim to server: ",posted_claim_to_server
		#DBGOUT#print "posted claim to entangled: ",posted_claim_to_entangled
		if self.transport.lower() == 'entangled':
			server = self.entangled_server
		else:
			server = self.transport
		twisted.internet.reactor.stop()
		
	def run_get_key_announcement(self):
		if self.fetch_from_entangled == True:
			server = self.entangled_server
			isEntangled = True
		else:
			server = self.transport
			isEntangled = False
		if re_keyid_ka.match(self.get_keyids[0]): # key id case
			self.keyannounce.get_key_announcements(server,None,self.get_keyids,isEntangled,self.get_key_announcement_callback,self.timeout)
		else:
			self.keyannounce.get_key_announcements(server,self.get_keyids[0],None,isEntangled,self.get_key_announcement_callback,self.timeout)

	def run_push_messages(self):
		self.poster.start_post_message(self.post_messages_callback)

	def run_fetch_messages(self):
		self.fetcher.get_new_messages(self.client_keyid,self.run_fetch_messages_callback)
	
	def run_fetch_messages_callback(self):
		#DBGOUT#print "Done fetching messages"
		twisted.internet.reactor.stop()

	def get_key_announcement_callback(self,announcements,claims):
		#DBGOUT#print "Get key announcement done, nr =",len(announcements),len(claims)
		for an in announcements:
			keyid,content = an
			#DBGOUT#print "KEYID ",keyid
			for l in content:
				#DBGOUT#print l
			print
			self.local_store.storeList(keyid,content)
			found,announcement_text = self.local_store.retrieve(keyid)
			result,fingerprint,age,cause,gpgstatus = self.keyannounce.verify_key_announcement_message(announcement_text,0,0)
			#DBGOUT#print "Announcement found and saved,",fingerprint
			#DBGOUT#print "Verify result",result,cause
		twisted.internet.reactor.stop()

	def post_messages_callback(self):
		#DBGOUT#print "Done posting messages"
		twisted.internet.reactor.stop()
		

	def status_callback(self,message):
		#DBGOUT#print "Status:",message

cliobj = cli()
cliobj.main(sys.argv)
sys.exit(0)

# EOF
